import pytest
import pandas as pd

@pytest.fixture
def geolocation_rules():
    """
    Returns the geolocation rules as a Pandas DataFrame
    """
    geolocation_rules_file = "augur/data/geolocation_rules.tsv"
    return pd.read_csv(
        geolocation_rules_file,
        sep="\t",
        names=['raw', 'annotations'],
        comment='#')


def test_geolocation_rules_are_not_duplicated(geolocation_rules):
    """
    Check for duplicated raw values.
    We want to avoid overwriting rules within our own rules.
    """
    duplicate_rule = geolocation_rules.loc[geolocation_rules.raw.duplicated()]
    assert len(duplicate_rule) == 0


def test_geolocation_rules_are_changing_values(geolocation_rules):
    """
    Check for rules where the raw values and annotations are the same.
    These rules are unnecessary as the values will be unchanged.
    """
    unchanged_rule = geolocation_rules.loc[geolocation_rules.raw == geolocation_rules.annotations]
    assert len(unchanged_rule) == 0


def test_geolocation_rules_are_not_cyclic(geolocation_rules):
    """
    Check for rules where the annotations exist in the raw values and
    the raw values exist in the annotations since this is the simplest
    way to flag cyclic rules.

    It won't catch all cyclic rules because wildcards might mask them,
    but it's better than nothing!
    """
    # Filter rules to ones that are changing values to make sure the same rules
    # don't get flagged by this test and `test_geolocation_rules_are_changing_values`
    changed_rules = geolocation_rules.loc[geolocation_rules.raw != geolocation_rules.annotations]
    cyclic_rules = changed_rules.loc[
        changed_rules.annotations.isin(changed_rules.raw) &
        changed_rules.raw.isin(changed_rules.annotations)
    ]
    assert len(cyclic_rules) == 0



import argparse
import datetime
from textwrap import dedent
import isodate
import pandas as pd
import re
from treetime.utils import numeric_date as tt_numeric_date, datetime_from_numeric
from typing import Any, Dict, Optional, Tuple, Union
from augur.errors import AugurError
from .errors import InvalidDate

from .ambiguous_date import AmbiguousDate

SUPPORTED_DATE_HELP_TEXT = dedent("""\
    1. an Augur-style numeric date with the year as the integer part (e.g. 2020.42) or
    2. a date in ISO 8601 date format (i.e. YYYY-MM-DD) (e.g. '2020-06-04') or
    3. a backwards-looking relative date in ISO 8601 duration format with optional P prefix (e.g. '1W', 'P1W')
""")

def date_to_numeric(date: datetime.date) -> float:
    """Wrapper around treetime.utils.numeric_date that ensures a float is returned."""
    value = tt_numeric_date(date)
    if not isinstance(value, float):
        raise Exception("treetime.utils.numeric_date unexpectedly returned a non-float.")
    return value

def numeric_date(date):
    """
    Converts the given *date* string to a :py:class:`float`.

    *date* may be given as:
    1. A string or float (number) with year as the integer part
    2. A string in the YYYY-MM-DD (ISO 8601) syntax
    3. A string representing a relative date (duration before datetime.date.today())

    Examples
    --------
    >>> numeric_date("2020.42")
    2020.42
    >>> numeric_date("2020-06-04")
    2020.42486...
    >>> import datetime, isodate
    >>> numeric_date("1W") == date_to_numeric(datetime.date.today() - isodate.parse_duration("P1W"))
    True
    """
    # date is a datetime.date
    if isinstance(date, datetime.date):
        return date_to_numeric(date)

    # date is numeric
    try:
        return float(date)
    except ValueError:
        pass

    # date is in YYYY-MM-DD form
    try:
        return date_to_numeric(datetime.date(*map(int, date.split("-", 2))))
    except ValueError:
        pass

    # date is a duration treated as a backwards-looking relative date
    try:
        # make a copy of date for this block
        duration_str = str(date)
        if duration_str.startswith('P'):
            duration_str = duration_str
        else:
            duration_str = 'P'+duration_str
        return date_to_numeric(datetime.date.today() - isodate.parse_duration(duration_str))
    except (ValueError, isodate.ISO8601Error):
        pass

    raise InvalidDate(date, f"""Ensure it is in one of the supported formats:\n{SUPPORTED_DATE_HELP_TEXT}""")

def numeric_date_type(date):
    """Wraps numeric_date() for argparse usage.

    This raises an ArgumentTypeError from InvalidDateFormat exceptions, otherwise the custom exception message won't be shown in console output due to:
    https://github.com/python/cpython/blob/5c4d1f6e0e192653560ae2941a6677fbf4fbd1f2/Lib/argparse.py#L2503-L2513
    """
    try:
        return numeric_date(date)
    except InvalidDate as error:
        raise argparse.ArgumentTypeError(str(error)) from error

def is_date_ambiguous(date, ambiguous_by):
    """
    Returns whether a given date string in the format of YYYY-MM-DD is ambiguous by a given part of the date (e.g., day, month, year, or any parts).

    Parameters
    ----------
    date : str
        Date string in the format of YYYY-MM-DD
    ambiguous_by : str
        Field of the date string to test for ambiguity ("day", "month", "year", "any")
    """
    date_components = date.split('-', 2)

    if len(date_components) == 3:
        year, month, day = date_components
    elif len(date_components) == 2:
        year, month = date_components
        day = "XX"
    else:
        year = date_components[0] if date_components[0] else 'X'
        month = "XX"
        day = "XX"

    # Determine ambiguity hierarchically such that, for example, an ambiguous
    # month implicates an ambiguous day even when day information is available.
    return any((
        "X" in year,
        "X" in month and ambiguous_by in ("any", "month", "day"),
        "X" in day and ambiguous_by in ("any", "day")
    ))

RE_NUMERIC_DATE = re.compile(r'^-*\d+\.\d+$')
"""
Matches floats (e.g. 2018.0, -2018.0).
Note that a year-only value is treated as incomplete ambiguous and must be
non-negative (see :const:`RE_YEAR_ONLY`).
"""

RE_YEAR_ONLY = re.compile(r'^\d+$')
"""
Matches:

1. Incomplete ambiguous ISO 8601 dates that are missing both the month and day
   parts (e.g. 2018)
2. Other positive integers (e.g. 1, 123, 12345)
"""

RE_YEAR_MONTH_ONLY = re.compile(r'^\d+-\d+$')
"""
Matches:

1. Reduced precision ISO 8601 dates that are missing the day part (e.g. 2018-03)
2. Other dates in <year>-<month> format (e.g. 2018-3)

Note: This matches out of bounds dates such as 2018-13.
Those should be further validated by date conversion functions.
"""

RE_YEAR_MONTH_DAY = re.compile(r'^\d+-\d+-\d+$')
"""
Matches:

1. ISO 8601 dates (e.g. 2018-03-09)
2. Other dates in <year>-<month>-<day> format (e.g. 2018-3-9)

Note: This matches out of bounds dates such as 2018-03-32.
Those should be further validated by date conversion functions.
"""

RE_AUGUR_AMBIGUOUS_DATE = re.compile(r'.*XX.*')
"""
Matches an Augur-style ambiguous date with 'XX' used to mask unknown parts of the date.
Note that this can support any date format, not just YYYY-MM-DD.
"""

def get_numerical_date_from_value(value, fmt, min_max_year=None) -> Union[float, Tuple[float, float], None]:
    value = str(value)

    # 1. Check if value is an exact date in the specified format (fmt).

    try:
        return date_to_numeric(datetime.datetime.strptime(value, fmt))
    except:
        pass
    
    # 2. Check if value is an ambiguous date in the specified format (fmt).

    if RE_AUGUR_AMBIGUOUS_DATE.match(value):
        start, end = AmbiguousDate(value, fmt=fmt).range(min_max_year=min_max_year)
        return (date_to_numeric(start), date_to_numeric(end))

    # 3. Check formats that are always supported.

    if RE_NUMERIC_DATE.match(value):
        return float(value)

    if RE_YEAR_ONLY.match(value):
        start, end = AmbiguousDate(f"{value}-XX-XX", fmt="%Y-%m-%d").range(min_max_year=min_max_year)
        return (date_to_numeric(start), date_to_numeric(end))

    if RE_YEAR_MONTH_ONLY.match(value):
        start, end = AmbiguousDate(f"{value}-XX", fmt="%Y-%m-%d").range(min_max_year=min_max_year)
        return (date_to_numeric(start), date_to_numeric(end))

    if RE_YEAR_MONTH_DAY.match(value):
        try:
            return date_to_numeric(datetime.date(*map(int, value.split('-'))))
        except ValueError as error:
            # Note: This isn't consistent with the out of bounds handling in
            # AmbiguousDate, which auto-converts out of bounds values to the
            # closest in-bound value.
            raise InvalidDate(value, str(error)) from error

    # 4. Return none (silent error) if the date does not match any of the checked formats.

    return None

def get_numerical_dates(
    metadata: pd.DataFrame,
    fmt,
    name_col = None,
    date_col = 'date',
    min_max_year = None,
) -> Dict[str, Union[float, Tuple[float, float], None]]:
    if not isinstance(metadata, pd.DataFrame):
        raise AugurError("Metadata should be a pandas.DataFrame.")

    strains = metadata.index.values
    dates = metadata[date_col].apply(
        lambda date: get_numerical_date_from_value(
            date,
            fmt,
            min_max_year
        )
    ).values

    return dict(zip(strains, dates))

def get_year_month(year, month):
    return f"{year}-{str(month).zfill(2)}"

def get_year_week(year, month, day):
    year, week = datetime.date(year, month, day).isocalendar()[:2]
    return f"{year}-{str(week).zfill(2)}"

def get_year_month_day(value: Any) -> Tuple[Optional[int], Optional[int], Optional[int]]:
    """
    Extract year, month, and day from a date value.

    Individual date components can be None if unresolvable.
    """
    date_or_range = get_numerical_date_from_value(value, fmt="%Y-%m-%d")

    if date_or_range is None:
        return (None, None, None)

    # Date range
    elif isinstance(date_or_range, tuple):
        start, end = (datetime_from_numeric(date_or_range[0]),
                      datetime_from_numeric(date_or_range[1]))

        # Only use unambiguous date components
        year = start.year if start.year == end.year else None
        month = start.month if start.month == end.month else None
        day = start.day if start.day == end.day else None

        return (year, month, day)

    # Exact date
    else:
        dt = datetime_from_numeric(date_or_range)
        return (dt.year, dt.month, dt.day)



"""
Refine an initial tree using sequence metadata.
"""
import numpy as np
import sys
from Bio import Phylo
from textwrap import dedent
from .argparse_ import ExtendOverwriteDefault
from .dates import get_numerical_dates
from .dates.errors import InvalidYearBounds
from .io.metadata import DEFAULT_DELIMITERS, DEFAULT_ID_COLUMNS, METADATA_DATE_COLUMN, InvalidDelimiter, Metadata, read_metadata
from .utils import read_tree, write_json, InvalidTreeError
from .errors import AugurError
from treetime.vcf_utils import read_vcf
from treetime.seq_utils import profile_maps

def refine(tree=None, aln=None, ref=None, dates=None, branch_length_inference='auto',
             confidence=False, resolve_polytomies=True, stochastic_resolve=False, max_iter=2, precision='auto',
             infer_gtr=True, Tc=0.01, reroot=None, use_marginal='always', fixed_pi=None, use_fft=True,
             clock_rate=None, clock_std=None, clock_filter_iqd=None, verbosity=1, covariance=True, rng_seed=None, **kwarks):
    from treetime import TreeTime

    try: #Tc could be a number or  'opt' or 'skyline'. TreeTime expects a float or int if a number.
        Tc = float(Tc)
    except ValueError:
        True #let it remain a string

    if (ref is not None) and (fixed_pi is None): #if VCF, fix pi
        #Otherwise mutation TO gaps is overestimated b/c of seq length
        fixed_pi = [ref.count(base)/len(ref) for base in ['A','C','G','T','-']]
        if fixed_pi[-1] == 0:
            fixed_pi[-1] = 0.05
            fixed_pi = [v-0.01 for v in fixed_pi]

    if ref is not None: # VCF -> adjust branch length
        #set branch length mode explicitly if auto, as informative-site only
        #trees can have big branch lengths, making this set incorrectly in TreeTime
        if branch_length_inference == 'auto':
            branch_length_inference = 'joint'

    #send ref, if is None, does no harm
    tt = TreeTime(tree=tree, aln=aln, ref=ref, dates=dates, use_fft=use_fft,
                  verbose=verbosity, gtr='JC69', precision=precision, rng_seed=rng_seed)

    # conditionally run clock-filter and remove bad tips
    if clock_filter_iqd:
        # treetime clock filter will mark, but not remove bad tips
        tt.clock_filter(reroot=reroot, n_iqd=clock_filter_iqd, plot=False) #use whatever was specified
        # remove them explicitly
        leaves = [x for x in tt.tree.get_terminals()]
        for n in leaves:
            if n.bad_branch:
                tt.tree.prune(n)
                print('pruning leaf ', n.name)
        # fix treetime set-up for new tree topology
        tt.prepare_tree()

    if confidence and use_marginal:
        # estimate confidence intervals via marginal ML and assign
        # marginal ML times to nodes
        marginal = 'always' if use_fft else 'assign'
    else:
        marginal = 'only-final' if confidence else False

    # uncertainty of the the clock rate is relevant if confidence intervals are estimated
    if confidence and clock_std:
        vary_rate = clock_std # if standard devivation of clock is specified, use that
    elif (clock_rate is None) and confidence and covariance:
        vary_rate = True      # if run in covariance mode, standard deviation can be estimated
    else:
        vary_rate = False     # otherwise, rate uncertainty will be ignored

    tt.run(infer_gtr=infer_gtr, root=reroot, Tc=Tc, time_marginal=marginal,
           branch_length_mode=branch_length_inference, resolve_polytomies=resolve_polytomies,
           stochastic_resolve=stochastic_resolve, max_iter=max_iter, fixed_pi=fixed_pi,
           fixed_clock_rate=clock_rate, vary_rate=vary_rate, use_covariation=covariance,
           raise_uncaught_exceptions=True, **kwarks)

    if confidence:
        for n in tt.tree.find_clades():
            n.num_date_confidence = list(tt.get_max_posterior_region(n, 0.9))

    print("\nInferred a time resolved phylogeny using TreeTime:"
          "\n\tSagulenko et al. TreeTime: Maximum-likelihood phylodynamic analysis"
          "\n\tVirus Evolution, vol 4, https://academic.oup.com/ve/article/4/1/vex042/4794731\n")
    return tt


def collect_node_data(T, attributes):
    data = {}
    for n in T.find_clades():
        data[n.name] = {
            attr: getattr(n, attr)
            for attr in attributes
            if getattr(n, attr, None) is not None
        }
    return data

class InvalidUseOfRemoveOutgroup(AugurError):
    """
    Raised when --remove-outgroup is used with an invalid --root option.
    """
    def __init__(self, root):
        self.root = root
    def __str__(self):
        return f"--remove-outgroup is not valid with {self.root!r} rooting - you must supply a single strain (outgroup) to root the tree"


def root_outside_of_treetime(T, root, is_timetree, remove_outgroup):
    """
    If the requested root(ing approach) is possible to do without TreeTime then this function
    modifies `T` in place, and returns `None`. If the rooting approach requires TreeTime then
    we return a string to be supplied to TreeTime as its root/reroot arguments.
    """
    treetime_reroot = None
    assert isinstance(root, list), "Internal augur refine error (requested root should be a list)"
    if root[0] in ['best', 'least-squares', 'oldest', 'min_dev'] and len(root)==1:
        if remove_outgroup:
            raise InvalidUseOfRemoveOutgroup(root[0])
        if is_timetree:
            treetime_reroot = root[0]
        else:
            if root[0]=='best': # CLI default, so print a warning rather than an error
                print("Warning: To root without inferring a timetree, you must specify an explicit outgroup.")
                print("\tProceeding without re-rooting. To suppress this message, use '--keep-root'.\n")
            else:
                raise AugurError(f"The rooting option {root[0]!r} is only available when inferring a timetree. Please specify an explicit outgroup or 'mid_point'.")
    elif root[0]=="mid_point" and len(root)==1:
        if remove_outgroup:
            raise InvalidUseOfRemoveOutgroup(root[0])
        T.root_at_midpoint()
    else:
        try:
            T.root_with_outgroup(root)
        except ValueError as err:
            tip_names = T.get_terminals()
            missing_names = [name for name in root if name not in tip_names]
            if len(missing_names):
                raise AugurError(dedent(f"""\
                    Rooting with the provided strain name(s) failed.
                    This is probably because the following names supplied to '--root' were not found in the tree:
                      - {f'{chr(10)}{" "*22}- '.join([repr(name) for name in missing_names])}
                    """))
            else:
                # raising a ValueError from the original ValueError will print the original message as well
                raise ValueError("Rooting with the provided strain name(s) failed for an unknown reason") from err

        if remove_outgroup:
            if len(root)>1:
                # run this check after attempting to root so that (more likely) errors such as misspelling
                # (e.g. "midpoint") are already caught & we know the rooting was successful
                raise InvalidUseOfRemoveOutgroup('multiple strains')
            try:
                outgroup = [node for node in T.root.clades if node.name==root[0]][0]
            except IndexError:
                raise IndexError(f"Internal error: outgroup {root[0]!r} was not a child of the tree root node")
            T.prune(outgroup)
            T.root.branch_length = 0.0

    return treetime_reroot

def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("refine", help=__doc__)
    parser.add_argument('--alignment', '-a', help="alignment in fasta or VCF format")
    parser.add_argument('--tree', '-t', required=True, help="prebuilt Newick")
    parser.add_argument('--metadata', type=str, metavar="FILE", help="sequence metadata")
    parser.add_argument('--metadata-delimiters', default=DEFAULT_DELIMITERS, nargs="+", action=ExtendOverwriteDefault,
                        help="delimiters to accept when reading a metadata file. Only one delimiter will be inferred.")
    parser.add_argument('--metadata-id-columns', default=DEFAULT_ID_COLUMNS, nargs="+", action=ExtendOverwriteDefault,
                        help="names of possible metadata columns containing identifier information, ordered by priority. Only one ID column will be inferred.")
    parser.add_argument('--output-tree', type=str, help='file name to write tree to')
    parser.add_argument('--output-node-data', type=str, help='file name to write branch lengths as node data')
    parser.add_argument('--use-fft', action="store_true", help="produce timetree using FFT for convolutions")
    parser.add_argument('--max-iter', default=2, type=int, help="maximal number of iterations TreeTime uses for timetree inference")
    parser.add_argument('--timetree', action="store_true", help="produce timetree using treetime, requires tree where branch length is in units of average number of nucleotide or protein substitutions per site (and branch lengths do not exceed 4)")
    parser.add_argument('--coalescent', help="coalescent time scale in units of inverse clock rate (float), optimize as scalar ('opt'), or skyline ('skyline')")
    parser.add_argument('--gen-per-year', default=50, type=float, help="number of generations per year, relevant for skyline output('skyline')")
    parser.add_argument('--clock-rate', type=float, help="fixed clock rate")
    parser.add_argument('--clock-std-dev', type=float, help="standard deviation of the fixed clock_rate estimate")
    parser.add_argument('--root', nargs="+", action=ExtendOverwriteDefault, default=['best'], help="rooting mechanism ('best', least-squares', 'min_dev', 'oldest', 'mid_point') "
                                "OR node to root by OR two nodes indicating a monophyletic group to root by. "
                                "Run treetime -h for definitions of rooting methods.")
    parser.add_argument('--keep-root', action="store_true", help="do not reroot the tree; use it as-is. "
                                "Overrides anything specified by --root.")
    parser.add_argument('--remove-outgroup', action="store_true", help="Remove the outgroup supplied via '--root'"
                                "This is only valid when a single strain name has been supplied as the root.")
    parser.add_argument('--covariance', dest='covariance', action='store_true', help="Account for covariation when estimating "
                                "rates and/or rerooting. "
                                "Use --no-covariance to turn off.")
    parser.add_argument('--no-covariance', dest='covariance', action='store_false')  #If you set help here, it displays 'default: True' - which is confusing!

    resolve_group = parser.add_mutually_exclusive_group()
    resolve_group.add_argument('--keep-polytomies', action='store_true', help='Do not attempt to resolve polytomies')
    resolve_group.add_argument('--stochastic-resolve', action='store_true', help='Resolve polytomies via stochastic subtree building rather than greedy optimization')
    resolve_group.add_argument('--greedy-resolve', action='store_false', dest='stochastic_resolve') # inverse of `--stochastic-resolve` to facilitate changing defaults in the future

    parser.add_argument('--precision', type=int, choices=[0,1,2,3], help="precision used by TreeTime to determine the number of grid points that are used for the evaluation of the branch length interpolation objects. Values range from 0 (rough) to 3 (ultra fine) and default to 'auto'.")
    parser.add_argument('--date-format', default="%Y-%m-%d", help="date format")
    parser.add_argument('--date-confidence', action="store_true", help="calculate confidence intervals for node dates")
    parser.add_argument('--date-inference', default='joint', choices=["joint", "marginal"],
                                help="assign internal nodes to their marginally most likely dates, not jointly most likely")
    parser.add_argument('--branch-length-inference', default='auto', choices = ['auto', 'joint', 'marginal', 'input'],
                                help='branch length mode of treetime to use')
    parser.add_argument('--clock-filter-iqd', type=float, help='clock-filter: remove tips that deviate more than n_iqd '
                                'interquartile ranges from the root-to-tip vs time regression')
    parser.add_argument('--vcf-reference', type=str, help='fasta file of the sequence the VCF was mapped to')
    parser.add_argument('--year-bounds', type=int, nargs='+', action=ExtendOverwriteDefault, help='specify min or max & min prediction bounds for samples with XX in year')
    parser.add_argument('--divergence-units', type=str, choices=['mutations', 'mutations-per-site'],
                        default='mutations-per-site', help='Units in which sequence divergences is exported.')
    parser.add_argument('--seed', type=int, help='seed for random number generation')
    parser.add_argument('--verbosity', type=int, default=1, help='treetime verbosity, between 0 and 6 (higher values more output)')
    parser.set_defaults(covariance=True)
    return parser


def run(args):

    # check alignment type, set flags, read in if VCF
    is_vcf = False
    ref = None

    # node data is the dict that will be exported as json
    node_data = {'alignment': args.alignment}
    # list of node attributes that are to be exported, will grow
    attributes = ['branch_length', 'confidence']

    try:
        T = read_tree(args.tree)
        node_data['input_tree'] = args.tree
    except (FileNotFoundError, InvalidTreeError) as error:
        print("ERROR: %s" % error, file=sys.stderr)
        return 1

    if not args.alignment:
        if args.timetree:
            print("ERROR: alignment is required for ancestral reconstruction or timetree inference", file=sys.stderr)
            return 1

        if args.divergence_units=='mutations':
            print("ERROR: alignment is required for divergence in units of mutations", file=sys.stderr)
            return 1

        # fake alignment to appease treetime when only using it for naming nodes...
        from Bio import SeqRecord, Seq, Align
        seqs = []
        for n in T.get_terminals():
            seqs.append(SeqRecord.SeqRecord(seq=Seq.Seq('ACGT'), id=n.name, name=n.name, description=''))
        aln = Align.MultipleSeqAlignment(seqs)
    elif any([args.alignment.lower().endswith(x) for x in ['.vcf', '.vcf.gz']]):
        if not args.vcf_reference:
            print("ERROR: a reference Fasta is required with VCF-format alignments", file=sys.stderr)
            return 1

        compress_seq = read_vcf(args.alignment, args.vcf_reference)
        aln = compress_seq['sequences']
        ref = compress_seq['reference']
        is_vcf = True
    else:
        aln = args.alignment

    from treetime import version as treetime_version
    print(f"augur refine is using TreeTime version {treetime_version}")

    # if not specified, construct default output file name with suffix _tt.nwk
    if args.output_tree:
        tree_fname = args.output_tree
    elif args.alignment:
        tree_fname = '.'.join(args.alignment.split('.')[:-1]) + '_tt.nwk'
    else:
        tree_fname = '.'.join(args.tree.split('.')[:-1]) + '_tt.nwk'

    # Parse rooting command line arguments and, for all rooting requests which are possible to do
    # outside of treetime we do them now (and ensure treetime doesn't subsequently reroot)
    # Note: if we have `--keep-root` we ignore `--root` completely
    treetime_reroot = None if args.keep_root else root_outside_of_treetime(T, args.root, args.timetree, args.remove_outgroup)

    if args.timetree:
        # load meta data and covert dates to numeric
        if args.metadata is None:
            print("ERROR: meta data with dates is required for time tree reconstruction", file=sys.stderr)
            return 1

        try:
            metadata_object = Metadata(args.metadata, args.metadata_delimiters, args.metadata_id_columns)
        except InvalidDelimiter:
            raise AugurError(
                f"Could not determine the delimiter of {args.metadata!r}. "
                f"Valid delimiters are: {args.metadata_delimiters!r}. "
                "This can be changed with --metadata-delimiters."
            )

        metadata = read_metadata(
            args.metadata,
            delimiters=[metadata_object.delimiter],
            columns=[metadata_object.id_column, METADATA_DATE_COLUMN],
            id_columns=[metadata_object.id_column],
            dtype="string",
        )

        try:
            dates = get_numerical_dates(metadata, fmt=args.date_format,
                                        min_max_year=args.year_bounds)
        except InvalidYearBounds as error:
            raise AugurError(f"Invalid value for --year-bounds: {error}")

        # save input state string for later export
        for n in T.get_terminals():
            if n.name in metadata.index and METADATA_DATE_COLUMN in metadata.columns:
                n.raw_date = metadata.at[n.name, METADATA_DATE_COLUMN]

        if args.date_confidence:
            time_inference_mode = 'always' if args.date_inference=='marginal' else 'only-final'
        else:
            time_inference_mode = 'always' if args.date_inference=='marginal' else 'never'

        tt = refine(tree=T, aln=aln, ref=ref, dates=dates, confidence=args.date_confidence,
                    reroot=treetime_reroot,
                    Tc=0.01 if args.coalescent is None else args.coalescent, #use 0.01 as default coalescent time scale
                    use_marginal = time_inference_mode, use_fft=args.use_fft,
                    branch_length_inference = args.branch_length_inference or 'auto',
                    precision = 'auto' if args.precision is None else args.precision,
                    clock_rate=args.clock_rate, clock_std=args.clock_std_dev,
                    clock_filter_iqd=args.clock_filter_iqd, max_iter=args.max_iter,
                    covariance=args.covariance, resolve_polytomies=(not args.keep_polytomies),
                    stochastic_resolve=args.stochastic_resolve, verbosity=args.verbosity, rng_seed=args.seed)

        node_data['clock'] = {'rate': tt.date2dist.clock_rate,
                              'intercept': tt.date2dist.intercept,
                              'rtt_Tmrca': -tt.date2dist.intercept/tt.date2dist.clock_rate}
        # Include the standard deviation of the clock rate, if the covariance
        # matrix is available.
        if hasattr(tt.date2dist, "cov") and tt.date2dist.cov is not None:
            node_data["clock"]["cov"] = tt.date2dist.cov
            node_data["clock"]["rate_std"] = np.sqrt(tt.date2dist.cov[0, 0])

        if args.coalescent=='skyline':
            try:
                skyline, conf = tt.merger_model.skyline_inferred(gen=args.gen_per_year, confidence=2)
                node_data['skyline'] = [[float(x) for x in skyline.x], [float(y) for y in conf[0]],
                                        [float(y) for y in skyline.y], [float(y) for y in conf[1]]]
            except:
                print(
                    "ERROR: skyline optimization by TreeTime has failed.",
                    "To avoid this error, try running without coalescent optimization or with `--coalescent opt` instead of `--coalescent skyline`.",
                    file=sys.stderr
                )
                return 1

        for node in T.find_clades():
            inferred = node.raw_date_constraint is None or isinstance(node.raw_date_constraint, list)
            setattr(node, 'date_inferred', inferred)

        attributes.extend(['numdate', 'clock_length', 'mutation_length', 'raw_date', 'date', 'date_inferred'])
        if args.date_confidence:
            attributes.append('num_date_confidence')
    else:
        from treetime import TreeAnc
        # instantiate treetime for the sole reason to name internal nodes
        tt = TreeAnc(tree=T, aln=aln, ref=ref, gtr='JC69', verbose=args.verbosity, rng_seed=args.seed)

    node_data['nodes'] = collect_node_data(T, attributes)
    if args.divergence_units=='mutations-per-site': #default
        pass
    elif args.divergence_units=='mutations':
        # find column of the compressed alignment that maps to masked positions of all N
        compressed_alignment = np.array([x for x in tt.data.compressed_alignment.values()]).T
        ambiguous_positions = []
        for ci, col in enumerate(compressed_alignment):
            if np.unique(col).size==1 and col[0]==tt.gtr.ambiguous:
                # the column with all Ns has only one unique value, and the first value is N
                ambiguous_positions.extend(tt.data.compressed_to_full_sequence_map[ci])

                # Since TreeTime represents all columns of all Ns with a single
                # entry in its compressed alignment structure, we can stop
                # searching the data structure.
                break
        # convert to set for faster lookup
        ambiguous_positions = set(ambiguous_positions)

        if not args.timetree:
            # infer ancestral sequence for the purpose of counting mutations
            # sample mutations from the root profile, otherwise use most likely state.
            # Reconstruct tip states to avoid mutations to N or W etc be counted
            tt.infer_ancestral_sequences(marginal=True, reconstruct_tip_states=True, sample_from_profile='root')
        nuc_map = profile_maps['nuc']

        def are_sequence_states_different(nuc1, nuc2):
            '''
            determine whether two ancestral states should count as mutation for divergence estimates
            while correctly accounting for ambiguous nucleotides
            '''
            if nuc1 in ['-', 'N'] or nuc2 in ['-', 'N']:
                return False
            elif nuc1 in nuc_map and nuc2 in nuc_map:
                return np.sum(nuc_map[nuc1]*nuc_map[nuc2])==0
            else:
                return False

        for node in T.find_clades():
            n_muts = len([
                position
                for ancestral, position, derived in node.mutations
                if are_sequence_states_different(ancestral, derived) and position not in ambiguous_positions
            ])

            if args.timetree:
                node_data['nodes'][node.name]['mutation_length'] = n_muts

            node_data['nodes'][node.name]['branch_length'] = n_muts
    else:
        print("ERROR: divergence unit",args.divergence_units,"not supported!", file=sys.stderr)
        return 1

    # Export refined tree and node data
    tree_success = Phylo.write(T, tree_fname, 'newick', format_branch_length='%1.8f', branch_length_only=True)
    print("updated tree written to",tree_fname, file=sys.stdout)

    if args.output_node_data:
        node_data_fname = args.output_node_data
    elif args.alignment:
        node_data_fname = '.'.join(args.alignment.split('.')[:-1]) + '.node_data.json'
    else:
        node_data_fname = '.'.join(args.tree.split('.')[:-1]) + '.node_data.json'

    write_json(node_data, node_data_fname)
    print("node attributes written to",node_data_fname, file=sys.stdout)

    return 0 if tree_success else 1



# Originally based on tests/help.py from the Nextstrain CLI project.¹
#
# ¹ <https://github.com/nextstrain/cli/blob/4a00d7100eff811eab6df34db73c7f6d4196e22b/tests/help.py>
import pytest

from augur import make_parser
from augur.argparse_ import walk_commands, ExtendOverwriteDefault


# Walking commands is slow, so do it only once and use it for all tests in this
# file (though currently n=1).
commands = list(walk_commands(make_parser()))


# Ensure we always use ExtendOverwriteDefault for options that take a variable
# number of arguments.  See <https://github.com/nextstrain/augur/pull/1709>.
@pytest.mark.parametrize("action", [
    pytest.param(action, id = " ".join(command) + " " + "/".join(action.option_strings))
        for command, parser in commands
        for action in parser._actions
         if action.nargs in {"+", "*"}
])
def test_ExtendOverwriteDefault(action):
    assert isinstance(action, ExtendOverwriteDefault)



import warnings

# Set up warnings & exceptions
_deprecationWarningsEmitted = False

def deprecated(message):
    warnings.warn(message, DeprecationWarning, stacklevel=2)
    global _deprecationWarningsEmitted
    _deprecationWarningsEmitted=True

def warn(message):
    warnings.warn(message, UserWarning, stacklevel=2)

def configure_warnings():
    # This function must be run before any calls to `warn()` and `deprecated()`, however
    # don't execute the code here since that would apply this configuration to _all_
    # augur commands due to the way all commands are pulled in by the augur runner (augur/__init__.py)
    #
    # Intended usage: call this function at the top of the entry function for the augur (sub)command, e.g.
    #       def run(args):
    #           configure_warnings()
    #           ...
    def customformatwarning(message, category, filename, lineno, line=None):
        if category.__name__ == "UserWarning":
            return "WARNING: {}\n\n".format(message)
        if category.__name__ == "DeprecationWarning":
            return "DEPRECATED: {}\n\n".format(message)
        return "{}\n".format(message)

    warnings.formatwarning = customformatwarning
    warnings.simplefilter("default") # show DeprecationWarnings by default

def deprecationWarningsEmitted():
    return _deprecationWarningsEmitted


import sys
import os
import json
from ..io.file import open_file
from ..errors import AugurError
from .warnings import deprecated
from ..types import ValidationMode
from ..validate import auspice_config_v2 as validate_auspice_config_v2, ValidateError, validation_failure
from collections import defaultdict
from typing import Union, Any
from collections.abc import Callable


def read_json(fname):
    if not (fname and os.path.isfile(fname)):
        print("ERROR: config file %s not found."%fname)
        return defaultdict(dict)

    try:
        with open_file(fname, 'rb') as ifile:
            config = json.load(ifile)
    except json.decoder.JSONDecodeError as err:
        print("FATAL ERROR:")
        print("\tCouldn't parse the JSON file {}".format(fname))
        print("\tError message: '{}'".format(err.msg))
        print("\tLine number: '{}'".format(err.lineno))
        print("\tColumn number: '{}'".format(err.colno))
        print("\tYou must correct this file in order to proceed.")
        sys.exit(2)

    return config

def _merge_lists(base: dict[str, Any], overlay: dict[str, Any], config_key: str, identity_func: Callable) -> dict[str, Any]:
    """
    This merging approach is used for config keys which are encoded as lists
    such as maintainers, colorings, filters.

    Returns an extended list with elements from the base (config's) list and the
    overlay (config's) list. If elements in the overlay are present in the base
    list, as determined by the *identity_func* then the base element is replaced
    with the overlay element.

    NOTE: Currently there is no way to remove elements from the base list, nor
    is it possible to insert a new (overlay) element at a specific position.
    Both are possible but increase the conceptual and programmatic complexity.
    """

    if config_key not in overlay:
        return base
    if config_key not in base:
        return {**base, config_key: overlay[config_key]}

    if not isinstance(overlay[config_key], list) or not isinstance(base[config_key], list):
        raise AugurError(f"Config merging for {config_key!r} failed as one (or more) entries were not lists")

    values = [*base[config_key]]
    for overlay_el in overlay[config_key]:
        for (idx, base_el) in enumerate(values):
            if identity_func(overlay_el, base_el):
                values[idx] = overlay_el # replace existing element with one from the overlay config
                break
        else:
            values.append(overlay_el)

    return {**base, config_key: values}

def _merge_scalar(base: dict[str, Any], overlay: dict[str, Any], config_key: str) -> dict[str, Any]:
    return {**base, config_key: overlay[config_key]} if config_key in overlay else base

def _merge_dicts(base: dict[str, Any], overlay: dict[str, Any], config_key: str) -> dict[str, Any]:
    """
    Dicts are _not_ recursively merged, they are simply merged by adding new keys from the overlay to
    those keys already present in the base. Identical keys replace the original key.
    """
    if config_key not in overlay:
        return base
    if config_key not in base:
        return {**base, config_key: overlay[config_key]}

    if not isinstance(overlay[config_key], dict) or not isinstance(base[config_key], dict):
        raise AugurError(f"Config merging for {config_key!r} failed as one (or more) entries were not dictionaries")

    return {**base, config_key: {**base[config_key], **overlay[config_key]}}

def _geo_resolution_id(el: Union[str, dict[str,str]]) -> str:
    return el['key'] if isinstance(el, dict) else el

def _replace_deprecated(config: dict[str,Any], old_name: str, new_name: Union[None,str], modify: Union[None, Callable]=None):
    if old_name not in config:
        return # NO-OP
    elif new_name is None:
        deprecated(f"[config file] key {old_name!r} is no longer used and has been dropped from your config")
        del config[old_name]
    elif old_name == new_name: # indicates deprecations within an unchanged top-level key
        assert modify is not None, "[internal error] modify must be a function"
        config[new_name] = modify(config[old_name])
    elif new_name in config:
        deprecated(f"[config file] top level keys {new_name!r} and {old_name!r} were both present, " \
                   "however the latter is a deprecated version of the former and will be ignored.")
        del config[old_name]
    else:
        deprecated(f"[config file] top level key {old_name!r} is deprecated. Renaming to {new_name!r}" +
                   (" (with modifications)" if modify else ""))
        config[new_name] = modify(config[old_name]) if modify else config[old_name]
        del config[old_name]

def _parse_color_options(options: Any) -> list[dict]:
    # parse v1-style 'color_options' & convert to v2-style 'colorings'
    if not isinstance(options, dict):
        raise AugurError(f"[config file] The deprecated field 'color_options' must be a dictionary")
    colorings = []
    for key, info in options.items():
        # note: if both legendTitle & menuItem are present then we use the latter. See https://github.com/nextstrain/auspice/issues/730
        if "menuItem" in info:
            deprecated("[config file] coloring '{}': 'menuItem' has been replaced with 'title'. Using 'menuItem' as 'title'.".format(key))
            info["title"] = info["menuItem"]
            del info["menuItem"]
        if "legendTitle" in info:
            if "title" in info: # this can only have been set via menuItem ^^^
                deprecated("[config file] coloring '{}': 'legendTitle' has been replaced with 'title' & is unused since 'menuItem' is present.".format(key))
            else:
                deprecated("[config file] coloring '{}': 'legendTitle' has been replaced with 'title'. Using 'legendTitle' as 'title'.".format(key))
                info["title"] = info["legendTitle"]
            del info["legendTitle"]
        if "key" in info:
            del info["key"]
        if info.get("type") == "discrete":
            deprecated("[config file] coloring '{}': type 'discrete' is no longer valid. Please use 'ordinal', 'categorical' or 'boolean'. "
                "This has been automatically changed to 'categorical'.".format(key))
            info["type"] = "categorical"
        colorings.append({"key": key, **info})
    return colorings

def _rename_deprecated_authors_coloring(colorings: list):
    # The 'authors' coloring (plural) is now called 'author' (singular)
    author  = next(iter([{'index': idx, 'coloring': c} for [idx, c] in enumerate(colorings) if c['key']=='author'] or [None]))
    authors = next(iter([{'index': idx, 'coloring': c} for [idx, c] in enumerate(colorings) if c['key']=='authors'] or [None]))
    if authors:
        if author:
            deprecated("[config file] ignoring deprecated 'authors' because a coloring for 'author' already exists")
            colorings = [c for [idx, c] in enumerate(colorings) if idx!=authors['index']]
        else:
            deprecated("[config file] renaming coloring 'authors' to 'author'")
            colorings[authors['index']]['key'] = "author"
    return colorings

def _rename_display_keys(display: dict) -> dict:
    defaults = {**display}
    v1_v2_keys = [
        # [ <old key>, <new key> ]
        ["geoResolution", "geo_resolution"],
        ["colorBy", "color_by"],
        ["distanceMeasure", "distance_measure"],
        ["mapTriplicate", "map_triplicate"]
    ]
    for [v1_key, v2_key] in [x for x in v1_v2_keys if x[0] in defaults]:
        deprecated("[config file] '{}' has been replaced with '{}'".format(v1_key, v2_key))
        defaults[v2_key] = defaults[v1_key]
        del defaults[v1_key]
    return defaults

def remove_unused_metadata_columns(columns: list) -> list:
    # 1. Remove any occurrences of 'author' and 'num_date' as it's a no-op - author
    #    and numerical date information is always exported on nodes if it's available.
    # 2. Remove "authors" as that was historically corrected to "author" during parsing,
    #    and similarly for 'numdate'
    return [n for n in columns if n not in ['author', 'authors', 'num_date', 'numdate']]

DEPRECATIONS = [
    {"old_name": 'vaccine_choices', "new_name": None}, # removed from config
    {"old_name": "updated", "new_name": None}, # removed from config
    {"old_name": "defaults", "new_name": "display_defaults", "modify": _rename_display_keys},
    {"old_name": "maintainer", "new_name": "maintainers", "modify": lambda m: [{"name": m[0], "url": m[1]}]},
    {"old_name": "geo", "new_name": "geo_resolutions", "modify": lambda values: [{"key": v} for v in values]},
    {"old_name": "color_options", "new_name": "colorings", "modify": _parse_color_options},
    {"old_name": "colorings", "new_name": "colorings", "modify": _rename_deprecated_authors_coloring},
    {"old_name": "metadata_columns", "new_name": "metadata_columns", "modify": remove_unused_metadata_columns},
]


def read_single_auspice_config(fname: str, validation_mode: ValidationMode) -> dict[str, Any]:
    config = read_json(fname)

    if validation_mode is not ValidationMode.SKIP:
        try:
            print(f"Validating config file {fname!r} against the JSON schema")
            validate_auspice_config_v2(fname)
        except ValidateError:
            print(f"Validation of {fname!r} failed. Please check the formatting of this file & refer to the augur documentation for further help. ")
            validation_failure(validation_mode)

    for data in DEPRECATIONS:
        _replace_deprecated(config, **data) # type: ignore[arg-type]

    return config

def merge_configs(configs: list[dict[str, Any]], validation_mode: ValidationMode, output_fname: Union[str,None]=None) -> dict[str, Any]:
    merged = configs[0]
    for overlay in configs[1:]:
        # We could leverage the config schemas here in the future if desired
        merged = _merge_scalar(merged, overlay, 'title')
        merged = _merge_lists(merged, overlay, 'colorings', lambda x,y: x.get('key')==y.get('key'))
        merged = _merge_lists(merged, overlay, 'geo_resolutions', lambda x,y: _geo_resolution_id(x)==_geo_resolution_id(y))
        merged = _merge_lists(merged, overlay, 'maintainers', lambda x,y: x.get('name')==y.get('name'))
        merged = _merge_scalar(merged, overlay, 'build_url')
        merged = _merge_scalar(merged, overlay, 'build_avatar')
        merged = _merge_lists(merged, overlay, 'filters', lambda x,y: x==y)
        merged = _merge_dicts(merged, overlay, 'display_defaults')
        merged = _merge_lists(merged, overlay, 'panels', lambda x,y: x==y)
        merged = _merge_lists(merged, overlay, 'data_provenance', lambda x,y: x.get('name')==y.get('name'))
        merged = _merge_lists(merged, overlay, 'metadata_columns', lambda x,y: x==y)

        # extensions have any type (as per the schema), but in practice I've only seen dicts used.
        # We merge these by taking the first encountered extension block, and if it's a dict
        # then any future extension blocks which are also dicts are merged in
        if 'extensions' not in merged and 'extensions' in overlay:
            merged['extensions'] = overlay['extensions']
        elif isinstance(merged.get('extensions', None), dict) and isinstance(overlay.get('extensions', None), dict):
            merged = _merge_dicts(merged, overlay, 'extensions')

    # Write the merged config before validation (if requested)
    if output_fname:
        print(f"Writing merged auspice config JSON to {output_fname!r}")
        # don't use our util function `write_json` as we don't want any modification of values
        with open_file(output_fname, 'w', encoding='utf-8') as handle:
            json.dump(merged, handle, indent=2)

    if validation_mode is not ValidationMode.SKIP:
        try:
            print(f"Validating merged config file against the JSON schema")
            validate_auspice_config_v2(merged)
        except ValidateError:
            if output_fname:
                print(f"Validation of merged config file failed. The merged JSON has been written to {output_fname!r} for debugging purposes", file=sys.stderr)
            else:
                print(f"Validation of merged config file failed. Here is the contents of the merged config JSON for debugging purposes:", file=sys.stderr)
                json.dump(merged, sys.stderr, indent=2)
            validation_failure(validation_mode)

    return merged


def read_auspice_configs(*fnames: str, validation_mode: ValidationMode, output_fname: Union[str,None]) -> dict[str, Any]:
    """
    Parses zero, one or multiple files as auspice config JSONs and returns a (merged) config dict.
    """
    configs = [read_single_auspice_config(fname, validation_mode) for fname in fnames]

    if len(configs)==0: # no config JSONs provided
        return {}
    elif len(configs)==1: # no merging necessary
        return configs[0]
    return merge_configs(configs, validation_mode, output_fname)
